module ELPOEKBO where

import Data.List
import Term
import Rule
import TRS
import SMT
import Algebra
import qualified ELPO
import qualified EKBO

data Parameter = ELPO ELPO.Parameter | EKBO EKBO.Parameter

instance Show Parameter where
  show (ELPO param) = ELPO.showParameter param
  show (EKBO param) = EKBO.showParameter param

varELPO :: Formula
varELPO = FVar "elpo"

varGt :: TRS -> Rule -> Formula
varGt trs rule
  | Just i <- elemIndex rule trs = FVar ("gt" ++ show i)
  | otherwise                    = error "varGt"

encode :: Signature -> TRS -> (Formula, [(Rule, Formula)])
encode sig trs =
  (conj (Distinct [ varPrec f | f <- fs ] :
         [ Geq (varA f) (Val 0) | f <- fs ] ++
         [ iff (ELPO.varGtLPO a s t) (ELPO.encodeGtLPO a s t) | (s, t) <- a ] ++
         [ iff (ELPO.varGtELPO a s t) (ELPO.encodeGtELPO a s t) | (s, t) <- trs ] ++
         [ disj [varELPO, 
                 conj [EKBO.encodeValidityOfWeight sig, 
                       EKBO.encodeAdmissibility sig]] ] ++
         [ iff (EKBO.varGtEKBO trs l r) (EKBO.encodeGtEKBO l r) | (l, r) <- trs ] ++
         [ disj [neg (varGt trs (l, r)), 
                 conj [varELPO, ELPO.varGtELPO a l r],
                 conj [neg varELPO, EKBO.varGtEKBO trs l r]]
         | (l, r) <- trs ] 
        ),
   [ ((l, r), varGt trs (l, r)) | (l, r) <- trs ])
  where 
    a = nub [ (s, t) | (l, r) <- trs, s <- subterms l, t <- subterms r ]
    fs = TRS.functions trs

-- Decoding functions.

decode :: Signature -> Model -> Parameter
decode sig model
  | SMT.evalFormula model varELPO = ELPO (ELPO.decode sig model) 
  | otherwise                     = EKBO (EKBO.decode sig model) 

-- Comparison

gt :: Parameter -> Term -> Term -> Bool
gt (ELPO param) s t = ELPO.gt param s t
gt (EKBO param) s t = EKBO.gt param s t
