module LPO where

import Data.List
import Term
import Rule
import TRS
import SMT
import Algebra (Precedence, showPrecedence)

-- Pretty printers

showParameter :: Precedence -> String
showParameter p =
  "LPO with precedence:\n\n" ++
  showPrecedence p ++ "\n"

-- LPO

gtLex :: Eq a => (a -> a -> Bool) -> [a] -> [a] -> Bool
gtLex _ [] []              = False
gtLex rel (x : xs) (y : ys) = rel x y || (x == y && gtLex rel xs ys)
gtLex _ _ _ = error "gtLex"

gtPrec :: Precedence -> String -> String -> Bool
gtPrec [] _ _ = False
gtPrec (f : fs) g h
  | f == g    = elem h fs
  | otherwise = gtPrec fs g h 

gt :: Precedence -> Term -> Term -> Bool
gt prec s t = gt12 prec s t || gt3 prec s t

gt12 :: Precedence -> Term -> Term -> Bool
gt12 prec s@(F f ss) (F g ts) =
  (gtPrec prec f g ||
   (f == g && gtLex (gt prec) ss ts)) &&
  all (gt prec s) ts
gt12 _ _ _  = False
  
gt3 :: Precedence -> Term -> Term -> Bool
gt3 param (F _ ss) t = any (\s -> geq param s t) ss
gt3 _ _ _            = False

geq :: Precedence -> Term -> Term -> Bool
geq param s t = s == t || gt param s t

-- Encoding functions.

idPrec :: String -> String
idPrec f = "p" ++ f

varPrec :: String -> Exp
varPrec f = Var (idPrec f)

encodeGtPrec :: String -> String -> Formula
encodeGtPrec f g = Gt (varPrec f) (varPrec g)

varGtLPO :: TRS -> Term -> Term -> Formula
varGtLPO a s t
  | Just n <- elemIndex (s, t) a = FVar ("gtLPO" ++ show n)
  | otherwise                    = error "varGtLPO"

encodeGeqLPO :: TRS -> Term -> Term -> Formula
encodeGeqLPO a s t
  | s == t    = top
  | otherwise = varGtLPO a s t

encodeGtLPO :: TRS -> Term -> Term -> Formula
encodeGtLPO a s@(F _ _) t = disj [encodeGtLPO1 a s t, encodeGtLPO23 a s t]
encodeGtLPO _ _         _ = bottom

encodeGtLPO1 :: TRS -> Term -> Term -> Formula
encodeGtLPO1 a (F _ ss) t = disj [ encodeGeqLPO a s t | s <- ss ]
encodeGtLPO1 _ _        _ = bottom

encodeGtLPO23 :: TRS -> Term -> Term -> Formula
encodeGtLPO23 a s@(F f ss) (F g ts)
  | f == g    = encodeLexLPO a s ss ts
  | otherwise = conj (encodeGtPrec f g : [ varGtLPO a s tj | tj <- ts ])
encodeGtLPO23 _ _ _ = bottom

encodeLexLPO :: TRS -> Term -> [Term] -> [Term] -> Formula
encodeLexLPO _ _ [] [] = Or []
encodeLexLPO a u (s : ss) (t : ts)
  | s == t    = encodeLexLPO a u ss ts
  | otherwise = conj (varGtLPO a s t : [ varGtLPO a u v | v <- ts ])
encodeLexLPO _ _ _ _ = error "encodeLexLPO"

encode :: Signature -> TRS -> (Formula, [(Rule, Formula)])
encode sig trs =
  (conj (Distinct [ varPrec f | (f, _) <- sig ] :
         [ iff (varGtLPO a s t) (encodeGtLPO a s t) | (s, t) <- a ]),
   [ ((l, r), varGtLPO a l r) | (l, r) <- trs ])
  where
    a = nub [ (s, t) | (l, r) <- trs, s <- subterms l, t <- subterms r ]

-- Decoding functions.

decode :: Signature -> Model -> Precedence
decode sig model =
  [ f | (_, f) <- sort [ (- evalExp model (varPrec f), f) | (f, _) <- sig ] ]

