module LPOEKBO where

import Data.List
import Term
import Rule
import TRS
import SMT
import Algebra
import qualified LPO
import qualified EKBO

data Parameter = LPO Precedence | EKBO EKBO.Parameter

instance Show Parameter where
  show (LPO param)  = show param
  show (EKBO param) = show param

varLPO :: Formula
varLPO = FVar "lpo"

varGt :: TRS -> Rule -> Formula
varGt trs rule
  | Just i <- elemIndex rule trs = FVar ("gt" ++ show i)
  | otherwise                    = error "varGt"

encode :: Signature -> TRS -> (Formula, [(Rule, Formula)])
encode sig  trs =
  (conj (Distinct [ varPrec f | f <- fs ] :
         [ Geq (varA f) (Val 0) | f <- fs ] ++
         [ iff (LPO.varGtLPO a s t) (LPO.encodeGtLPO a s t) | (s, t) <- a ] ++
         [ iff (LPO.varGtLPO a s t) (LPO.encodeGtLPO a s t) | (s, t) <- trs ] ++
         [ disj [varLPO, 
                 conj [EKBO.encodeValidityOfWeight sig, 
                       EKBO.encodeAdmissibility sig]] ] ++
         [ iff (EKBO.varGtEKBO trs l r) (EKBO.encodeGtEKBO l r) | (l, r) <- trs ] ++
         [ disj [neg (varGt trs (l, r)), 
                 conj [varLPO, LPO.varGtLPO a l r],
                 conj [neg varLPO, EKBO.varGtEKBO trs l r]]
         | (l, r) <- trs ] 
        ),
   [ ((l, r), varGt trs (l, r)) | (l, r) <- trs ])
  where 
    a = nub [ (s, t) | (l, r) <- trs, s <- subterms l, t <- subterms r ]
    fs = TRS.functions trs

-- Decoding functions.

decode :: Signature -> Model -> Parameter
decode sig model
  | SMT.evalFormula model varLPO = LPO  (LPO.decode  sig model) 
  | otherwise                    = EKBO (EKBO.decode sig model) 

-- Comparison

gt :: Parameter -> Term -> Term -> Bool
gt (LPO  param) s t = LPO.gt param s t
gt (EKBO param) s t = EKBO.gt param s t
